// skein.cpp written and placed in the public domain by Jean-Pierre Muench

#include "pch.h"
#include "skein.h"
#include "Threefish.h"

NAMESPACE_BEGIN(CryptoPP)

Skein::Skein(const unsigned int DigestSize,unsigned int BlockSize) :
	m_DigestSize(DigestSize),m_Blocksize(BlockSize)
{
	if(m_Blocksize!=32 && m_Blocksize!=64 && m_Blocksize!=128)
		m_Blocksize=0;

	if(!m_Blocksize)
	{
		// use Threefish-256 only for Digests < 256 bits
		// use Threefish-512 for Digests d with 256 <= d <= 512
		// use Threefish-1024 for the rest
		if(m_DigestSize<32)
			m_Blocksize=32;
		else if(m_DigestSize<=64)
			m_Blocksize=64;
		else
			m_Blocksize=128;
	}

	switch (m_Blocksize)
	{
	case 32:
		m_Threefish.reset(new Threefish_256::Encryption());
		break;
	case 64:
		m_Threefish.reset(new Threefish_512::Encryption());
		break;
	case 128:
		m_Threefish.reset(new Threefish_1024::Encryption());
		break;
	}

	Restart();
}

void Skein::Update(const byte* input, size_t length)
{
	if(m_BufferForNextBlock.SizeInBytes() + length <= m_Blocksize)
	{
		if(m_BufferForNextBlock.SizeInBytes() + length <= m_BufferForNextBlock.SizeInBytes()) // counter buffer-/integeroverflows
			return;
		const size_t OldSize = m_BufferForNextBlock.SizeInBytes();
		m_BufferForNextBlock.CleanGrow(OldSize+length);
		memcpy(&m_BufferForNextBlock.BytePtr()[OldSize],input,length);

		return;
	}

	// process buffered data first
	SecByteBlock Buffer(m_Blocksize);
	memcpy(Buffer,m_BufferForNextBlock,m_BufferForNextBlock.SizeInBytes());
	const size_t ToCopy = m_Blocksize - m_BufferForNextBlock.SizeInBytes();
	memcpy(&Buffer.BytePtr()[m_BufferForNextBlock.SizeInBytes()],input,m_Blocksize - m_BufferForNextBlock.SizeInBytes());
	input += m_Blocksize - m_BufferForNextBlock.SizeInBytes();
	length -= m_Blocksize - m_BufferForNextBlock.SizeInBytes();

	word32 TweakFlags=0;
	TweakFlags |= 48ui32 <<24; // msg flag
	if(m_LowerCounter==0&&m_UpperCounter==0)
		TweakFlags |= 1ui32 << 30; // first block flag

	IncrementCounters(m_Blocksize);
	FixedSizeSecBlock<byte,16> Tweak;
	memset_z(Tweak,0,16);
	memcpy(&Tweak.BytePtr()[0],(const byte*)&m_LowerCounter,sizeof(m_LowerCounter));
	memcpy(&Tweak.BytePtr()[8],(const byte*)&m_UpperCounter,sizeof(m_UpperCounter));
	memcpy(&Tweak.BytePtr()[12],(const byte*)&TweakFlags,sizeof(TweakFlags));
	/*ConditionalByteReverse(LITTLE_ENDIAN_ORDER,);
	ConditionalByteReverse(LITTLE_ENDIAN_ORDER,);
	ConditionalByteReverse(LITTLE_ENDIAN_ORDER,);*/

	m_Threefish->SetKey(m_State,m_Blocksize);
	m_Threefish->ProcessAndXorBlockWithTweak(Buffer,Buffer,m_State,Tweak,16);

	while(length > m_Blocksize)
	{
		word32 Flags=0;
		Flags |= 48ui32 <<24; // msg flag

		IncrementCounters(m_Blocksize);
		FixedSizeSecBlock<byte,16> NewTweak;
		memset_z(NewTweak,0,16);
		memcpy(&NewTweak.BytePtr()[0],(const byte*)&m_LowerCounter,sizeof(m_LowerCounter));
		memcpy(&NewTweak.BytePtr()[8],(const byte*)&m_UpperCounter,sizeof(m_UpperCounter));
		memcpy(&NewTweak.BytePtr()[12],(const byte*)&Flags,sizeof(Flags));
		/*ConditionalByteReverse(LITTLE_ENDIAN_ORDER,&NewTweak.BytePtr()[0],(const byte*)&m_LowerCounter,sizeof(m_LowerCounter));
		ConditionalByteReverse(LITTLE_ENDIAN_ORDER,&NewTweak.BytePtr()[8],(const byte*)&m_UpperCounter,sizeof(m_UpperCounter));
		ConditionalByteReverse(LITTLE_ENDIAN_ORDER,&NewTweak.BytePtr()[12],(const byte*)&Flags,sizeof(Flags));*/

		m_Threefish->SetKey(m_State,m_Blocksize);
		m_Threefish->ProcessAndXorBlockWithTweak(input,input,m_State,NewTweak,16);

		input += m_Blocksize;
		length -= m_Blocksize;
	}

	m_BufferForNextBlock.CleanNew(length);
	memcpy(m_BufferForNextBlock,input,length);
}

void Skein::TruncatedFinal(byte *hash, size_t size)
{
	ThrowIfInvalidTruncatedSize(size);

	SecByteBlock LastBlock(m_Blocksize);
	memset_z(LastBlock,0,m_Blocksize);
	memcpy_s(LastBlock,m_Blocksize,m_BufferForNextBlock,m_BufferForNextBlock.SizeInBytes());

	word32 TweakFlags=0;
	TweakFlags |= 48ui32 <<24; // msg flag
	TweakFlags |= 1ui32 << 31; // last block flag
	if(m_LowerCounter==0&&m_UpperCounter==0)
		TweakFlags |= 1ui32 << 30; // first block flag

	IncrementCounters(m_BufferForNextBlock.SizeInBytes());
	m_BufferForNextBlock.CleanNew(0);

	FixedSizeSecBlock<byte,16> Tweak;
	memset_z(Tweak,0,16);
	memcpy(&Tweak.BytePtr()[0],(const byte*)&m_LowerCounter,sizeof(m_LowerCounter));
	memcpy(&Tweak.BytePtr()[8],(const byte*)&m_UpperCounter,sizeof(m_UpperCounter));
	memcpy(&Tweak.BytePtr()[12],(const byte*)&TweakFlags,sizeof(TweakFlags));

	m_Threefish->SetKey(m_State,m_Blocksize);
	m_Threefish->ProcessAndXorBlockWithTweak(LastBlock,LastBlock,m_State,Tweak,16);

	// all blocks are now processed
	// output transformation now

	m_Threefish->SetKey(m_State,m_Blocksize);
	word32 OutputTweakFlags=0;
	OutputTweakFlags |= 63ui32 <<24; // out flag
	OutputTweakFlags |= 1ui32 << 30; // first block flag
	OutputTweakFlags |= 1ui32 << 31; // last block flag
	FixedSizeSecBlock<byte,16> OutTweak;
	memset_z(OutTweak,0,16);
	
	SecByteBlock OutputBlock(m_Blocksize);
	word64 OutputCounter = 0;

	const word32 OutUBISize = sizeof(OutputCounter);
	memcpy(&OutTweak.BytePtr()[0],(const byte*)&OutUBISize,sizeof(OutUBISize));
	memcpy(&OutTweak.BytePtr()[12],(const byte*)&OutputTweakFlags,sizeof(OutputTweakFlags));

	for(word64 i=0;i<(size/m_Blocksize);++i)
	{
		memset_z(OutputBlock,0,m_Blocksize);
		ConditionalByteReverse(LITTLE_ENDIAN_ORDER,OutputBlock.BytePtr(),(const byte*)&OutputCounter,sizeof(OutputCounter));

		m_Threefish->ProcessAndXorBlockWithTweak(OutputBlock,NULL,hash + OutputCounter*m_Blocksize,OutTweak,16);
		++OutputCounter;
	}
	SecByteBlock LastDigestBlockBuffer(m_Blocksize);

	memset_z(OutputBlock,0,m_Blocksize);
	ConditionalByteReverse(LITTLE_ENDIAN_ORDER,OutputBlock.BytePtr(),(const byte*)&OutputCounter,sizeof(OutputCounter));

	m_Threefish->ProcessAndXorBlockWithTweak(OutputBlock,NULL,LastDigestBlockBuffer,OutTweak,16);
	memcpy(hash + OutputCounter*m_Blocksize,LastDigestBlockBuffer,size-OutputCounter*m_Blocksize);

	Restart();
}

void Skein::Restart()
{
	m_BufferForNextBlock.CleanNew(0);

	memset_z(m_State,0,128);
	FixedSizeSecBlock<byte,128> ConfigurationString;
	memset_z(ConfigurationString,0,m_Blocksize);
	const word32 SchemaIdentifier = 0x33414853; // = "SHA3"
	const word16 Version = 1;
	const word64 DigestSize = 8*m_DigestSize; // digest size in bits

	memcpy(&ConfigurationString.BytePtr()[0],(const byte*)&SchemaIdentifier,sizeof(SchemaIdentifier));
	memcpy(&ConfigurationString.BytePtr()[4],(const byte*)&Version,sizeof(Version));
	memcpy(&ConfigurationString.BytePtr()[8],(const byte*)&DigestSize,sizeof(DigestSize));

	word32 ByteCountForTweak = 32;
	//ByteCountForTweak=ConditionalByteReverse(LITTLE_ENDIAN_ORDER,ByteCountForTweak);
	word32 Flags = 0;
	Flags |= 1ui32 << 30; // first block in UBI computation
	Flags |= 1ui32 << 31; // last block in UBI computation
	Flags |= 4ui32 << 24; // config block in UBI computation

	FixedSizeSecBlock<byte,16> Tweak;
	memset_z(Tweak,0,16);
	memcpy(&Tweak.BytePtr()[0],(const byte*)&ByteCountForTweak,sizeof(ByteCountForTweak));
	memcpy(&Tweak.BytePtr()[12],(const byte*)&Flags,sizeof(Flags));

	m_Threefish->SetKey(m_State,m_Blocksize);
	m_Threefish->ProcessAndXorBlockWithTweak(ConfigurationString,ConfigurationString,m_State,Tweak,16);

	m_LowerCounter=0;
	m_UpperCounter=0;
}

void Skein::IncrementCounters(size_t n)
{
	const word64 OldLower = m_LowerCounter;
	const word32 OldUpper = m_UpperCounter;
	m_LowerCounter+=n;
	if(m_LowerCounter<OldLower)
		m_UpperCounter++;
	if(m_UpperCounter<OldUpper)
		throw(InvalidArgument("input string too long (longer than 96-bits)"));
}

NAMESPACE_END